# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Scan Genius Interface
"""

from __future__ import unicode_literals

import re
import os.path
import datetime
import socket
import shutil
import logging
from decimal import Decimal

import edbob

from rattail.core import Object
from rattail.csvutil import UnicodeReader, UnicodeDictReader
from rattail.gpc import GPC
from rattail_livnat.exceptions import UnexpectedDataFormat


log = logging.getLogger(__name__)


def collect_order(path):
    """
    The purpose of this function is to gather orders exported from Scan Genius
    on various machines, and move them to another folder so that they may
    become part of a single processing queue.
    """

    queue = edbob.config.require('rattail.sw.livnat', 'orders.collection_dir')
    order = parse_order_header(path)
    dest = os.path.join(queue, '%s %s %s.csv' % (
            socket.gethostname(),
            edbob.local_time().strftime('%Y-%m-%d_%H-%M-%S'),
            order.id))
    shutil.move(path, dest)


def collect_order_new(path):
    """
    The purpose of this function is to gather orders exported from Scan Genius
    on various machines, and move them to another folder so that they may
    become part of a single processing queue.
    """
    # We can't do anything useful with zero-byte files; just delete them.
    if os.path.getsize(path) == 0:
        log.info("deleting zero-byte file: {0}".format(path))
        os.remove(path)
        return

    queue = edbob.config.require('rattail.sw.livnat', 'orders.collection_dir')
    order = parse_order_header_new(path)
    dest = os.path.join(queue, '%s %s %s.csv' % (
            socket.gethostname(),
            edbob.local_time().strftime('%Y-%m-%d_%H-%M-%S'),
            order.id))
    shutil.move(path, dest)


class Order(Object):
    """
    Represents an order as parsed from a Scan Genius order export file.

    .. todo::
       Still need to go over the column definitions of the SG export, etc...
    """

    csv_path = None

    id = None
    date = None
    # fixme1 = None
    submitted = None
    store_name = None
    store_address1 = None
    store_address2 = None
    store_city = None
    store_state = None
    store_country = None
    store_zipcode = None
    account_number = None
    vendor_name = None
    vendor_phone = None

    total_cost = None

    def __init__(self, **kwargs):
        kwargs.setdefault('details', [])
        super(Order, self).__init__(**kwargs)

    def __repr__(self):
        return "<Order: %s, %s>" % (self.id, self.vendor_name)

    def parse_details(self):
        """
        Parses the details of an order file as exported from Scan Genius.  All
        :class:`OrderDetail` objects which are discovered from parsing will
        henceforth be available within the :attr:`Order.details` attribute of
        the ``Order`` instance.
        """

        del self.details[:]
        self.total_cost = Decimal('0.00')

        csv_file = open(self.csv_path, 'rb')
        reader = UnicodeReader(csv_file, encoding='latin_1')
        reader.next() # skip header line

        for line, row in enumerate(reader, 2):
            good = False
            if len(row) == 12:
                good = True

            # Apparently SG doesn't always include the suggested retail field.
            elif len(row) == 11:
                row.insert(3, '')
                good = True

            # One time I found a record with two missing "stock" fields...?
            # See https://redmine.mamajeansmarket.com/issues/1147 for info.
            elif len(row) == 10:
                row.extend(['0', '0'])
                good = True

            if not good:
                raise UnexpectedDataFormat(self.csv_path, line)

            det = OrderDetail(line_number=line)
            det.upc = GPC(row[0])
            det.desc1 = row[1]
            det.desc2 = row[2]
            if row[3]:
                det.suggested_retail = Decimal(row[3])
            det.sku = row[4]
            det.unit_cost = Decimal(row[5])
            det.quantity = int(row[6])
            det.extended_cost = Decimal(row[7])
            det.case_quantity = int(row[8])
            det.stock_front = int(row[9])
            det.stock_back = int(row[10])
            det.stock_ordered = int(row[11])

            self.details.append(det)
            self.total_cost += det.extended_cost

        csv_file.close()

    def parse_details_new(self):
        """
        Parses the details of an order file as exported from Scan Genius.  All
        :class:`OrderDetail` objects which are discovered from parsing will
        henceforth be available within the :attr:`Order.details` attribute of
        the ``Order`` instance.
        """

        del self.details[:]
        self.total_cost = Decimal('0.00')

        with open(self.csv_path, 'rb') as csv_file:
            
            # We want to ignore the header section here.  However the use of
            # UnicodeDictReader below requires iteration, and yet also expects
            # the "beginning" of the file to contain the fieldnames.  To force
            # this all to play nicely, we calculate our offset and then seek to
            # it explicitly.  It's probably likely that there is a better way,
            # but this works.
            offset = self.find_details_offset(csv_file)
            csv_file.seek(offset)

            reader = UnicodeDictReader(csv_file, encoding='latin_1')
            for line, row in enumerate(reader, 2):

                detail = OrderDetail(line_number=line)
                detail.upc = GPC(row['UPC'])
                detail.desc1 = row['Product Name']
                if row['SRP']:
                    detail.suggested_retail = Decimal(row['SRP'])
                detail.sku = row['SKU']
                detail.unit_cost = Decimal(row['Price'])
                detail.quantity = int(row['QTY'])
                detail.extended_cost = Decimal(row['Total'])
                detail.case_quantity = int(row['CPS'])
                self.details.append(detail)
                self.total_cost += detail.extended_cost

    def find_details_offset(self, csv_file):
        """
        Find the character offset of the details data within a CSV file.
        """
        # First we must determine how many lines within the file comprise the
        # header data section.  This is 2 in most cases, but not always.  In
        # particular, the "Memo" field may contain line breaks.
        header_line_count = None
        reader = UnicodeReader(csv_file, encoding='latin_1')
        for i, row in enumerate(reader):
            if i == 1:
                header_line_count = reader.reader.line_num
                break
        if header_line_count is None:
            raise Exception("File header could not be accurately parsed: "
                            "{0}".format(csv_file.name))

        # Next we calculate the offset of the details section, based on the sum
        # character length of the header section.
        csv_file.seek(0)
        offset = 0
        for line, row in enumerate(csv_file, 1):
            offset += len(row)
            if line == header_line_count:
                break
        return offset


class OrderDetail(Object):
    """
    Represents a detail line within a :class:`Order` instance, as parsed from
    the export file.
    """

    line_number = None

    upc = None
    desc1 = None
    desc2 = None
    suggested_retail = None
    sku = None
    unit_cost = None
    quantity = None
    extended_cost = None
    case_quantity = None
    stock_front = None
    stock_back = None
    stock_ordered = None

    def __repr__(self):
        return "<OrderDetail: %s, %s>" % (self.line_number, self.desc1)


def parse_order_header(csv_path):
    """
    Reads the header record from ``csv_path`` and returns a populated
    :class:`Order` instance.
    """

    csv_file = open(csv_path, 'rb')
    csv_reader = UnicodeReader(csv_file, encoding='latin_1')
    csv_rec = csv_reader.next()
    csv_file.close()

    if len(csv_rec) != 14:
        raise UnexpectedDataFormat(csv_path, 1)

    order = Order(csv_path=csv_path)

    order.id = int(csv_rec[0])
    words = csv_rec[1].split()
    if not len(words) in (1, 3):
        raise UnexpectedDataFormat(csv_path, 1)
    order.date = datetime.datetime.strptime(words[0], '%m/%d/%Y').date()
    # order.fixme1 = csv_rec[2]
    if csv_rec[3]:
        order.submitted = datetime.datetime.strptime(csv_rec[3], '%m/%d/%Y %I:%M:%S %p')
    else:
        order.submitted = None
    order.store_name = csv_rec[4]
    order.store_address1 = csv_rec[5]
    order.store_address2 = csv_rec[6]
    order.store_city = csv_rec[7]
    order.store_state = csv_rec[8]
    order.store_country = csv_rec[9]
    order.store_zipcode = csv_rec[10]
    order.account_number = csv_rec[11]
    order.vendor_name = csv_rec[12]
    order.vendor_phone = csv_rec[13]
    
    order.exported = os.path.getmtime(csv_path)
    order.exported = datetime.datetime.fromtimestamp(order.exported)
    
    return order


def parse_order_header_new(csv_path):
    """
    Reads the header record from ``csv_path`` and returns a populated
    :class:`Order` instance.
    """
    with open(csv_path, 'rb') as csv_file:
        reader = UnicodeDictReader(csv_file, encoding='latin_1')
        header = reader.next()

    modtime = datetime.datetime.fromtimestamp(
        os.path.getmtime(csv_path))

    order = Order(csv_path=csv_path)
    order.id = header['PO']
    order.date = parse_date(header['OrderDate']) or modtime.date()
    order.store_name = header['Store']
    order.store_address1 = header['Address']
    order.store_city = header['City']
    order.store_state = header['State']
    order.store_zipcode = header['Zip']
    order.account_number = header['Acct Nbr']
    order.vendor_name = header['Supplier']
    order.exported = modtime
    return order


def parse_datetime(value):
    """
    Parse a ``datetime`` instance from the given string value.
    """
    # The date/time format seems to have changed over time; therefore we try
    # all of the known formats.  They are listed here with most recent on top.
    formats = [
        '%m/%d/%y %I:%M %p %Z',
        '%m/%d/%Y',
    ]
    for fmt in formats:
        try:
            return datetime.datetime.strptime(value, fmt)
        except ValueError:
            pass

    # As a special case, we'll try to remove the timezone info from the string,
    # if it is present.  Then we can try one more format which does not involve
    # the timezone.  Note that in practice this is thought to be the typical
    # logic path, due to http://bugs.python.org/issue22377
    value = re.sub(r' [A-Z]{3}$', '', value)
    try:
        return datetime.datetime.strptime(value, '%m/%d/%y %I:%M %p')
    except ValueError:
        pass


def parse_date(value):
    """
    Parse a date from the given string value.
    """
    dt = parse_datetime(value)
    if dt:
        return dt.date()
